from inspect import Signature
from typing import TYPE_CHECKING, Any, Dict, List, Set, Tuple, Type

from pydantic import create_model
from pydantic.fields import FieldInfo, Undefined
from pydantic_factories import ModelFactory
from typing_extensions import get_args

from starlite.exceptions import ImproperlyConfiguredException
from starlite.plugins.base import PluginMapping, PluginProtocol, get_plugin_for_value
from starlite.signature.models import PydanticSignatureModel, SignatureModel
from starlite.signature.parameter import SignatureParameter
from starlite.utils import is_dependency_field
from starlite.utils.helpers import unwrap_partial

if TYPE_CHECKING:
    from starlite.types import AnyCallable


def get_type_annotation_from_plugin(
    parameter: SignatureParameter, plugin: PluginProtocol, field_plugin_mappings: Dict[str, PluginMapping]
) -> Any:
    """Use plugin declared for parameter annotation type to generate a pydantic model.

    Args:
        parameter:  SignatureParameter
        plugin: PluginProtocol

    Returns:
        A pydantic model to be used as a type
    """
    type_args = get_args(parameter.annotation)
    type_value = type_args[0] if type_args else parameter.annotation
    field_plugin_mappings[parameter.name] = PluginMapping(plugin=plugin, model_class=type_value)
    pydantic_model = plugin.to_pydantic_model_class(model_class=type_value, parameter_name=parameter.name)
    if type_args:
        return List[pydantic_model]  # type:ignore[valid-type]
    return pydantic_model


def create_field_definition_from_parameter(parameter: SignatureParameter) -> Tuple[Any, Any]:
    """Construct an `(<annotation>, <default>)` tuple, appropriate for `pydantic.create_model()`.

    Args:
        parameter: SignatureParameter

    Returns:
        tuple[Any, Any]
    """
    if parameter.should_skip_validation:
        if is_dependency_field(parameter.default):
            return Any, parameter.default.default
        return Any, ...

    if ModelFactory.is_constrained_field(parameter.default):
        return parameter.default, ...

    if parameter.default_defined:
        return parameter.annotation, parameter.default

    if not parameter.optional:
        return parameter.annotation, ...

    return parameter.annotation, None


def parse_fn_signature(
    fn: "AnyCallable", plugins: List["PluginProtocol"], dependency_name_set: Set[str]
) -> Tuple[Dict[str, Any], Any, Dict[str, PluginMapping], Set[str]]:
    """Parse a function signature into data used for the generation of a signature model.

    Args:
        fn: A callable.
        plugins: A list of plugins.
        dependency_name_set: A set of dependency names

    Returns:
        A tuple containing the following values for generating a signature model: a mapping of field definitions, the
        callable's return annotation, a mapping of field names to plugins - if any, and an updated dependency name set.
    """

    signature = Signature.from_callable(fn)
    fn_name = getattr(fn, "__name__", "anonymous")

    field_plugin_mappings: Dict[str, PluginMapping] = {}
    field_definitions: Dict[str, Any] = {}
    dependency_names: Set[str] = set()
    for parameter in (
        SignatureParameter(fn_name, param_name, param_def)
        for param_name, param_def in signature.parameters.items()
        if param_name not in ("self", "cls")
    ):
        if is_dependency_field(parameter.default) and parameter.name not in dependency_name_set:
            if not parameter.optional and (
                isinstance(parameter.default, FieldInfo) and parameter.default.default is Undefined
            ):
                raise ImproperlyConfiguredException(
                    f"Explicit dependency '{parameter.name}' for '{fn_name}' has no default value, "
                    f"or provided dependency."
                )
            dependency_names.add(parameter.name)

        if plugin := get_plugin_for_value(value=parameter.annotation, plugins=plugins):
            parameter.annotation = get_type_annotation_from_plugin(parameter, plugin, field_plugin_mappings)

        field_definitions[parameter.name] = create_field_definition_from_parameter(parameter)

    return field_definitions, signature.return_annotation, field_plugin_mappings, dependency_names


def create_signature_model(
    fn: "AnyCallable", plugins: List["PluginProtocol"], dependency_name_set: Set[str]
) -> Type[SignatureModel]:
    """Create a model for a callable's signature. The model can than be used to parse and validate before passing it to
    the callable.

    Args:
        fn: A callable.
        plugins: A list of plugins.
        dependency_name_set: A set of dependency names

    Returns:
        A signature model.
    """
    unwrapped_fn = unwrap_partial(fn)
    fn_name = getattr(fn, "__name__", "anonymous")
    fn_module = getattr(fn, "__module__", None)

    field_definitions, return_annotation, field_plugin_mappings, dependency_names = parse_fn_signature(
        fn=unwrapped_fn,
        plugins=plugins,
        dependency_name_set=dependency_name_set,
    )

    model: Type[SignatureModel] = create_model(
        f"{fn_name}_signature_model",
        __base__=PydanticSignatureModel,
        __module__=fn_module or "pydantic.main",
        **field_definitions,
    )
    model.return_annotation = return_annotation
    model.field_plugin_mappings = field_plugin_mappings
    model.dependency_name_set = {*dependency_name_set, *dependency_names}
    return model



from abc import ABC, abstractmethod
from typing import Any, ClassVar, Dict, Set

from pydantic import BaseConfig, BaseModel, ValidationError

from starlite.connection import ASGIConnection, Request
from starlite.enums import ScopeType
from starlite.exceptions import InternalServerException, ValidationException
from starlite.plugins import PluginMapping


class SignatureModel(ABC):
    """Base model for Signature modelling."""

    dependency_name_set: ClassVar[Set[str]]
    field_plugin_mappings: ClassVar[Dict[str, PluginMapping]]
    return_annotation: ClassVar[Any]

    @classmethod
    @abstractmethod
    def parse_values_from_connection_kwargs(cls, connection: "ASGIConnection", **kwargs: Any) -> Dict[str, Any]:
        """Extract values from the connection instance and return a dict of parsed values.

        Args:
            connection: The ASGI connection instance.
            **kwargs: A dictionary of kwargs.

        Raises:
            ValidationException: If validation failed.
            InternalServerException: If another exception has been raised.

        Returns:
            A dictionary of parsed values
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def fields(cls) -> Dict[str, Any]:
        """Allow uniform access to the signature models fields, independent of the implementation.

        Returns:
            A string keyed mapping of field values.
        """
        raise NotImplementedError

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Normalize access to the signature model's dictionary method, because different backends use different methods
        for this.

        Returns: A dictionary of string keyed values.
        """
        raise NotImplementedError


class PydanticSignatureModel(SignatureModel, BaseModel):
    """Model that represents a function signature that uses a pydantic specific type or types."""

    class Config(BaseConfig):
        copy_on_model_validation = "none"
        arbitrary_types_allowed = True

    @classmethod
    def parse_values_from_connection_kwargs(cls, connection: "ASGIConnection", **kwargs: Any) -> Dict[str, Any]:
        """Extract values from the connection instance and return a dict of parsed values.

        Args:
            connection: The ASGI connection instance.
            **kwargs: A dictionary of kwargs.

        Raises:
            ValidationException: If validation failed.
            InternalServerException: If another exception has been raised.

        Returns:
            A dictionary of parsed values
        """
        try:
            signature = cls(**kwargs)
        except ValidationError as e:
            method = connection.method if isinstance(connection, Request) else ScopeType.WEBSOCKET
            if client_errors := [error for error in e.errors() if error["loc"][-1] not in cls.dependency_name_set]:
                raise ValidationException(
                    detail=f"Validation failed for {method} {connection.url}", extra=client_errors
                ) from e
            raise InternalServerException(
                detail=f"A dependency failed validation for {method} {connection.url}", extra=e.errors()
            ) from e

        return signature.to_dict()

    def _resolve_field_value(self, key: str) -> Any:
        """Return value using key mapping, if available.

        Args:
            key: A field name.

        Returns:
            The plugin value, if available.
        """
        value = self.__getattribute__(key)  # pylint: disable=unnecessary-dunder-call
        mapping = self.field_plugin_mappings.get(key)
        return mapping.get_model_instance_for_value(value) if mapping else value

    @classmethod
    def fields(cls) -> Dict[str, Any]:
        """Allow uniform access to the signature models fields, independent of the implementation.

        Returns:
            A string keyed mapping of field values.
        """
        return dict(cls.__fields__)

    def to_dict(self) -> Dict[str, Any]:
        """Normalize access to the signature model's dictionary method, because different backends use different methods
        for this.

        Returns: A dictionary of string keyed values.
        """
        if self.field_plugin_mappings:
            return {key: self._resolve_field_value(key) for key in self.__fields__}
        return {key: self.__getattribute__(key) for key in self.__fields__}  # pylint: disable=unnecessary-dunder-call



from .modelling_logic import create_signature_model
from .models import SignatureModel
from .utils import get_signature_model

__all__ = ("create_signature_model", "SignatureModel", "get_signature_model")



from inspect import Parameter, Signature
from typing import Any

from msgspec._core import nodefault as MsgspecUndefined
from pydantic.fields import Undefined as PydanticUndefined

from starlite.exceptions import ImproperlyConfiguredException
from starlite.utils import is_optional_union, should_skip_dependency_validation

UNDEFINED_SENTINELS = {PydanticUndefined, Signature.empty, MsgspecUndefined}
SKIP_VALIDATION_NAMES = {"request", "socket", "scope", "receive", "send"}


class SignatureParameter:
    """Represents the parameters of a callable for purpose of signature model generation."""

    __slots__ = (
        "annotation",
        "default",
        "name",
        "optional",
    )

    annotation: Any
    default: Any
    name: str
    optional: bool

    def __init__(self, fn_name: str, parameter_name: str, parameter: Parameter) -> None:
        """Initialize SignatureParameter.

        Args:
            fn_name: Name of function.
            parameter_name: Name of parameter.
            parameter: inspect.Parameter
        """
        if parameter.annotation is Signature.empty:
            raise ImproperlyConfiguredException(
                f"Kwarg {parameter_name} of {fn_name} does not have a type annotation. If it "
                f"should receive any value, use the 'Any' type."
            )
        self.annotation = parameter.annotation
        self.default = parameter.default
        self.name = parameter_name
        self.optional = is_optional_union(parameter.annotation)

    @property
    def default_defined(self) -> bool:
        """Whether a default value is defined for the parameter.

        Returns:
            A boolean determining whether a default value is defined.
        """
        return self.default not in UNDEFINED_SENTINELS

    @property
    def should_skip_validation(self) -> bool:
        """Whether the parameter should skip validation.

        Returns:
            A boolean indicating whether the parameter should be validated.
        """
        return self.name in SKIP_VALIDATION_NAMES or should_skip_dependency_validation(self.default)
